
/**
 * Title:        Appia<p>
 * Description:  Protocol development and composition framework<p>
 * Copyright:    Copyright (c) Alexandre Pinto & Hugo Miranda & Luis Rodrigues<p>
 * Company:      F.C.U.L.<p>
 * @author Alexandre Pinto & Hugo Miranda & Luis Rodrigues
 * @version 1.0
 */
package appia;

import java.util.Hashtable;

import appia.events.channel.*;

public class Channel {

	String channelID;
	private QoS qos;
	private Hashtable eventsRoutes=null;

	private EventScheduler eventScheduler;
	private TimerManager timerManager;

	private boolean alive=false;

	protected Session[] sessions=null;

	public Channel(String channelID, QoS qos, EventScheduler eventScheduler) {

		this.channelID=channelID;
		this.qos=qos;
		this.eventScheduler=eventScheduler;

 		sessions=new Session[qos.getLayers().length];

		timerManager=Appia.getTimerManager();
	}

	public String getChannelID() {
		return channelID;
	}

	public EventScheduler getEventScheduler() {
		return eventScheduler;
	}

	public ChannelEventRoute getEventRoute(Event event) throws AppiaEventException {
    if ( eventsRoutes == null )
       throw new AppiaEventException(AppiaEventException.CLOSEDCHANNEL,"Channel Not Started");

		ChannelEventRoute channelRoute=(ChannelEventRoute)eventsRoutes.get(event.getClass());

		if (channelRoute==null)
			 throw new AppiaEventException(AppiaEventException.UNWANTEDEVENT,"Unwanted Event");
		else
			return channelRoute;
	}

	public int getFirstSession(ChannelEventRoute channelRoute, Direction dir, Session source)
		throws AppiaEventException {

    Session[] route=channelRoute.getRoute();

		int i,index;

    if (source==null) {
      if (dir.direction==Direction.UP)
         return 0;
      else
         return route.length-1;
    }

    for(i=sessions.length-1 ; (i >= 0) && (sessions[i] != source) ; i--);

    if (i < 0)
       throw new AppiaEventException(AppiaEventException.UNKNOWNSESSION,"Session not memeber of Channel");

    i+=dir.direction;
    index=-1;
    while ( (i >= 0) && (i < sessions.length) && (index < 0) ) {
      for(index=route.length-1 ; (index >= 0) && (route[index] != sessions[i]) ; index--);
      i+=dir.direction;
    }

    return ((index < 0) ? route.length : index) ;
  }

/*
	public int getFirstSession(ChannelEventRoute channelRoute, Direction dir, Session source)
		throws appiaEventException {

    if ( sessions == null )
       throw new appiaEventException(appiaEventException.CLOSEDCHANNEL,"Channel Not Started");

		boolean[] waypoints=channelRoute.waypoints;
    Session[] route=channelRoute.getRoute();

		int i;

    if (source==null) {
      if (dir.direction==Direction.UP)
         return 0;
      else
         return route.length-1;
    }

    for (i=0 ; (i < sessions.length) && (sessions[i]!=source) ; i++) ;

		if (i >= sessions.length)
			throw new appiaEventException(appiaEventException.UNKNOWNSESSION,"Session not memeber of Channel");

    for ( ; (i < waypoints.length) && (i >= 0) && !waypoints[i] ; i+=dir.direction ) ;

    if ( (i >= waypoints.length) || (i < 0) )
       return route.length+1;

    if ( i == waypoints.length -1 )
       return route.length;

    int j;
    for ( j=0 ; (j < route.length) && (sessions[i] != route[j]) ; j++ ) ;
    return j;
	}
*/

	public QoS getQoS() {
		return qos;
	}

	public boolean equalQoS(Channel channel) {
		return this.qos==channel.qos;
	}

	public void start() throws AppiaDuplicatedSessionsException {
    synchronized (this) {
      alive=false;
    }

		int i;
		for (i=0 ; i < sessions.length ; i++) {
			if (sessions[i]!=null)
				sessions[i].boundSessions(this);
		}

		createUnboundedSessions();

		int j;
		for (i=0 ; i < sessions.length ; i++) {
			for (j=i+1 ; j < sessions.length ; j++) {
				if (sessions[i]==sessions[j])
					 throw new AppiaDuplicatedSessionsException();
			}
		}

		makeEventsRoutes();

    Appia.insertEventScheduler(eventScheduler);

		try {
      eventScheduler.insert(new ChannelInit(this));
    } catch (AppiaEventException e) {}
	}

	public void end() {
    synchronized (this) {
      if (alive) {
        alive=false;

        try {
          eventScheduler.insert(new ChannelClose(this));
        } catch (AppiaEventException e) {}
		  }
    }
	}

	public ChannelCursor getCursor() {
		return new ChannelCursor(this);
	}

  public void makeAsync() throws AppiaEventException {
    insertEvent(new AsyncEvent(this));
  }

	protected synchronized void insertEvent(Event event) throws AppiaEventException {
		if (!alive)
			 throw new AppiaEventException(AppiaEventException.CLOSEDCHANNEL,"Channel is Closed");

		eventScheduler.insert(event);
	}

	private void createUnboundedSessions() {
		int i;
		Layer[] layers=qos.getLayers();

		for (i=0 ; i < sessions.length ; i++) {
			if (sessions[i]==null)
				sessions[i]=layers[i].createSession();
		}
	}

	private void makeEventsRoutes() {
		QoSEventRoute[] qosRoutes=qos.getEventsRoutes();
    eventsRoutes=new Hashtable();

		for (int i=0 ; i < qosRoutes.length ; i++) {
			eventsRoutes.put(qosRoutes[i].getEventType(),qosRoutes[i].makeChannelRoute(this));
		}
	}

 	public void handle(Event event) {

    if ( event instanceof ChannelInit ) {
      synchronized (this) {
        alive=true;
      }
      return;
    }

    if ( event instanceof ChannelClose ) {
      eventsRoutes.clear();
		  eventScheduler=null;
		  timerManager=null;
      return;
    }

    if ( event instanceof Timer ) {
      timerManager.handleTimerRequest((Timer)event);
      return;
    }

    if ( event instanceof PeriodicTimer ) {
      timerManager.handlePeriodicTimer((PeriodicTimer)event);
      return;
    }

    if ( event instanceof AsyncEvent ) {
      return;
    }

    if ( event instanceof Debug ) {
      return;
    }

    if ( event instanceof EchoEvent ) {
      Event e=((EchoEvent)event).getEvent();

      e.setChannel(this);
      e.setDirection(event.getDirection());
      e.getDirection().invert();
      e.setSource(null);

      try {
        e.init();
        eventScheduler.insert(e);
      } catch (AppiaEventException ex) {
        System.err.println(
          "appia.Channel.handle(Event): exception (\""+
          ex.getMessage()+
          "\") while processing EchoEvent."
        );
      }
      return;
    }
  }
}