package appia.protocols.application;

//////////////////////////////////////////////////////////////////////
//                                                                  //
// Appia: protocol development and composition framework            //
//                                                                  //
// Version: 1.0/J                                                   //
//                                                                  //
// Copyright, 2000, Universidade de Lisboa                          //
// All rights reserved                                              //
// See license.txt for further information                          //
//                                                                  //
// Class: ApplicationSession: User interface with Appia             //
//                                                                  //
// Author: Hugo Miranda, 05/2000                                    //
//                                                                  //
// Change Log:                                                      //
// Data, alteracao, razao/consequncias                             //
//                                                                  //
//////////////////////////////////////////////////////////////////////

import appia.*;
import appia.events.*;
import appia.events.channel.*;
import appia.message.*;
import appia.protocols.udpsimple.RegisterSocketEvent;
import appia.protocols.udpsimple.InetWithPort;
import appia.protocols.fifo.FIFOUndelivered;

import java.io.*;
import java.util.*;
import java.net.InetAddress;

/**
 * Class ApplicationSession provides the dynamic behaviour for the 
 * simple user interface with Appia. This interface accepts a predefined
 * set of commands, forwarding them to Appia. It is mainly used for testing
 * @author Hugo Miranda
 * @see    ApplicationLayer
 * @see    Session
 */

public class ApplicationSession extends Session {

    private final int INVALIDCOMMAND=-1;
    private final int OPENSOCKET=0;
    private final int SENDMESSAGE=1;
    private final int HELP=2;
    private final int DEBUGMESSAGE=3;
    private final int TESTSTART=4;

    /* User IO */
    private PrintStream out=System.out;

    private boolean debug=false;
    
    /* Running test info */
    private TestEvent testModel;
    private long      startTime;
    private long      missingRetries;
    
    /* Appia */

    private Vector            channels;
    private QoS               myQoS;
    private int               myPort;
    private ApplicationReader reader;

    /**
     * Mainly used for corresponding layer initialization
     */

    public ApplicationSession(ApplicationLayer l) {
	super(l);

	channels=new Vector();
	this.myQoS=myQoS;
	this.myPort=myPort;
	reader=new ApplicationReader(this);
	reader.start();
    }

    public void init(QoS myQoS, int myPort) {
	this.myQoS=myQoS;
	this.myPort=myPort;
    }

    public Channel giveChannel() {
	return (Channel)channels.firstElement();
    }

    private Channel findChannel(String channelID) {
	Enumeration e=channels.elements();
	Channel     ret=null;

	while(ret==null && e.hasMoreElements()) {
	    Channel c=(Channel)(e.nextElement());
	    if(c.getChannelID().equals(channelID))
		ret=c;
	}

	return ret;
    }

    private int parseCommand(String command) {
	if(command.equals("open"))
	    return OPENSOCKET;
	if(command.equals("send"))
	    return SENDMESSAGE;
	if(command.equals("debug"))
	    return DEBUGMESSAGE;
	if(command.equals("help"))
	    return HELP;
	if(command.equals ("test"))
	    return TESTSTART;
	return INVALIDCOMMAND;
    }

    private void printHelp() {
	out.println("help\tprints this message");
	out.println("open channelid\tOpens a socket to a specific "+
		    "endpoint");
	out.println("send channelid adress port message\t Sends a "+
		    "message to a channel");
	out.println("debug channelid {start|stop|now} [file]");
	out.println("test channelid adress port tests-number");
    }

    private void bindSessions(Channel newChannel) {

	ChannelCursor cc=newChannel.getCursor();
	cc.top();

	/* This is the first channel: just bind this session */

	if(channels.isEmpty()) {
	    try {
		boolean found=false;

		while(!found && cc.isPositioned()) {
		    if(cc.getLayer() instanceof ApplicationLayer) {
			cc.setSession(this);
			found=true;
		    }
		    cc.down();
		}
		
		if(!found)
		    System.err.println("Couldn't find application "+
				       "session in QoS");
	    }
	    catch(AppiaCursorException ex) {
		System.err.println("Unexpected exception in Application "+
				   "Session");
	    }
	}

	else {

	    /* Other channels already exist. Copy the sessions */

	    ChannelCursor model=((Channel)channels.firstElement()).getCursor();
	    Session s=model.top();

	    try {
		while(cc.isPositioned()) {
		    cc.setSession(s);
		    cc.down();
		    s=model.down();
		}
		/* This one is tricky. isOnBottom returns true before we
		   set the last one */
		cc.setSession(s);
 	    }
	    catch(AppiaCursorException ex) {
		System.err.println("Unexpected exception in Application "+
				   "session");
	    }
	}
    }

    private void openSocket(StringTokenizer args) {
	if(args.hasMoreTokens()) {
	    String chName=args.nextToken();

	    /* Ensure that this channel doesn't exist yet */

	    if(findChannel(chName)!=null) {
		System.err.println("Channel already exists");
		return;
	    }

	    /* Asks the QoS to create a new channel */

	    Channel newChannel=myQoS.createUnboundChannel(chName);

	    /* Copy sessions from another channel (only one QoS exists) */

	    bindSessions(newChannel);

	    /* Start the channel */
	    try {
		newChannel.start();
	    }
	    catch(AppiaDuplicatedSessionsException ex) {
		System.err.println("Unexpected exception in Application "+
				   "session");
	    }
	}
	else
	    out.println("Channel name is missing");
    }

    private void sendDebug(StringTokenizer args) {
	if(args.countTokens()!=2 && args.countTokens()!=3) {
	    out.println("Invalid arguments number. Usage:");
	    out.println("debug channel, one of this options: start, "+
			"stop, now and an optional file name");
	    return;
	}
	
	Channel c=findChannel(args.nextToken());

	if(c==null) {
	    out.println("Channel not found. Create him first!");
	    return;
	}

	try {
	    EventQualifier eq=new EventQualifier();

	    String action=args.nextToken();
	    
	    if(action.equals("start"))
		eq.set(eq.ON);
	    else if(action.equals("stop"))
		eq.set(eq.OFF);
	    else if(action.equals("now"))
		eq.set(eq.NOTIFY);
	    else out.println("Invalid action. Use one of start, stop and now");
	    
	    OutputStream debugOut=out;

	    if(args.hasMoreTokens()) {
		try {
		    debugOut=new FileOutputStream(args.nextToken(),false);
		}
		catch(FileNotFoundException ex) {
		    out.println("File could not be oppened for output");
		    return;
		}
	    }

	    appia.events.channel.Debug e=
		new appia.events.channel.Debug(debugOut);
	    e.setChannel(c);
	    e.setDirection(new Direction(Direction.DOWN));
	    e.setSource(this);
	    e.setQualifier(eq);
	    
	    e.init();	    
	    e.go();
	}
	catch(appia.AppiaEventException ex) {
	    out.println("Unexpected exception when sending debug event");
	}
    }
	
    private void sendMessage(StringTokenizer args,String commandLine) {
	if(args.countTokens()<4) {
	    out.println("Invalid arguments number");
	    return;
	}

	Channel c=findChannel(args.nextToken());
	
	if(c==null) {
	    out.println("Channel not found. Create him first!");
	    return;
	}

	SendableEvent e=new ApplSendEvent();

	e.setDirection(new Direction(Direction.DOWN));
	e.setChannel(c);
	e.setSource(this);

	String dest=args.nextToken();
	String port=args.nextToken();

	try {
	    e.dest=new InetWithPort();
	    ((InetWithPort)e.dest).host=InetAddress.getByName(dest);
	    ((InetWithPort)e.dest).port=Integer.parseInt(port);

	    e.source=new InetWithPort();
	    ((InetWithPort)e.source).host=InetAddress.getLocalHost();
	    ((InetWithPort)e.source).port=-1;
	
	    String message=
		commandLine.substring(commandLine.indexOf(port)+
				      port.length()+1);

	    byte[] data=new byte[message.length()+4];
	    lengthToByte(data,message.length());
	    System.arraycopy(message.getBytes(),0,data,4,message.length());

	    e.getMessage().setByteArray(data,0,data.length);

	    e.init();
	    e.go();
	}
	catch(java.net.UnknownHostException ex) {
	    out.println("Unknown destination host");
	}
	catch(NumberFormatException ex) {
	    out.println("Port number must be an integer");
	}
	catch(appia.AppiaEventException ex) {
	    System.err.println("Unexpected exception in Application Session");
	    switch(ex.type) {
	    case appia.AppiaEventException.NOTINITIALIZED :
		System.err.println("Event not initialized in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.ATTRIBUTEMISSING :
		System.err.println("Missing attribute in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.UNKNOWNQUALIFIER :
		System.err.println("Unknown qualifier (impossible) in "+
				   " message sending (Application Session)");
		break;
	    case appia.AppiaEventException.UNKNOWNSESSION :
		System.err.println("Unknown session in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.UNWANTEDEVENT :
		System.err.println("Unwanted event in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.CLOSEDCHANNEL :
		System.err.println("Channel closed in message "+
				   "sending (Application Session)");
		break;
	    }
	}
    }

    /**
     * Parse and send user messages
     */

    private void parseUser() {
	synchronized(reader) {
	    if(reader.sComLine!=null) {
		StringTokenizer comLine=new StringTokenizer(reader.sComLine);
		
		if(comLine.hasMoreTokens()) {
		    String command=comLine.nextToken();
		    
		    switch(parseCommand(command)) {
		    case HELP :
			printHelp();
			break;
		    case OPENSOCKET :
			openSocket(comLine);
			break;
		    case SENDMESSAGE :
			sendMessage(comLine,reader.sComLine);
			break;
		    case DEBUGMESSAGE :
			sendDebug(comLine);
			break;
		    case TESTSTART :
			startTest(comLine);
			break;
		    default :
			System.out.println("Invalid command");
		    }
		}
		reader.sComLine=null;
		reader.notify();
	    }
	}
    }

    /**
     * This protocol copies all sessions from the first channel
     */

    public void boundSessions(Channel channel) {
	if(!channels.isEmpty()) {
	    ChannelCursor model=((Channel)channels.firstElement()).getCursor(),
		cc=channel.getCursor();
	    Session sModel=model.top();
	    Session sNew=cc.top();

	    try {
		while(cc.isPositioned()) {
		    if(sNew==null) {
			cc.setSession(sModel);
			System.out.println("Another session bounded");
		    }
		    sModel=model.down();
		    sNew=cc.down();
		}
	    }
	    catch(AppiaCursorException ex) {
		System.err.println("Unexpected excepcion in Application "+
				   "Session");
	    }
	}
    }

    /**
     * Main Event handler function. Accepts all incoming events and
     * dispatches them to the appropriate functions
     * @ params e The incoming event
     * @ see Session
     */

    public void handle(Event e) {
	if(e instanceof ChannelInit) {
	    handleChannelInit((ChannelInit)e);
	    parseUser();
	}
	else if(e instanceof ApplSendEvent) {
	    receiveData((SendableEvent)e);
	    parseUser();
	}
	else if(e instanceof TestEvent) {
	    replyTest((TestEvent)e);
	}
	else if(e instanceof ReplyTestEvent) {
	    handleTestReply((ReplyTestEvent)e);
	}
	else if(e instanceof FIFOUndelivered) {
	    handleUndelivered((FIFOUndelivered)e);
	    parseUser();
	}
	else {
	    parseUser();
	    try {
		e.go();
	    }
	    catch(AppiaEventException ex) {
		System.err.println("Unexpected exception in Application "+
				   "session");
	    }
	}
    }

    private void lengthToByte(byte[] header,int length) {
	long factor=(long)Math.pow(2,24);

	for(int i=3;i>=0;i--,factor/=256) {
	    header[i]=(byte)(length/factor);
	    length-=header[i]*factor;
	}
    }

    private int byteToLength(MsgBuffer buf) {
	int acum=0;
	long factor=1;
	int  off=buf.off;

	for(int i=0;i<=3;i++,factor*=256,off++) {
	    acum+=buf.data[off]*factor;
	}

	return acum;
    }

    private void receiveData(SendableEvent e) {

	/* Echoes received messages to the user */
	if(e.getDirection().direction==Direction.UP) {
	    Message m=e.getMessage();
	    MsgBuffer msgBuf=new MsgBuffer();
	    msgBuf.len=4;
	    m.pop(msgBuf);
	    int l=byteToLength(msgBuf);

	    out.println("Message with "+l+ " bytes received from "+
			((InetWithPort)e.source).host.getHostName() +
			" on channel "+e.getChannel().getChannelID()+
			" (port "+ ((InetWithPort)e.source).port + ") :");
	    while(l>0) {
		msgBuf.len=l;
		m.pop(msgBuf);
		out.print(new String(msgBuf.data,msgBuf.off,msgBuf.len));
		l-=msgBuf.len;
	    }
	    System.out.println();
	}
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception in Application "+
			       "session");
	}
    }
	    

    private void handleChannelInit(ChannelInit e) {

	/* Forwards channel init event. New events must follow this
	   one */
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception in Application "+
			       "session");
	}

	/* If this is the first channel. Informs layers bellow of the port
	   where messages will be received and sent */
       
	try {
	    if(channels.isEmpty()) {
		RegisterSocketEvent rse=new
		    RegisterSocketEvent(e.getChannel(),
					new Direction(Direction.DOWN),
					this,myPort);
		rse.go();
	    }
	}
	catch(appia.AppiaEventException ex) {
	    switch(ex.type) {
	    case appia.AppiaEventException.UNWANTEDEVENT :
		System.err.println("The QoS definition doesn't satisfy the "+
				   "application session needs. "+
				   "RegisterSocketEvent, received by "+
				   "UdpSimpleSession is not being acepted");
		break;
	    default :
		System.err.println("Unexpected exception in "+
				   "Application session");
		break;
	    }
	}

	/* Adds the channel to the vector */
	channels.add(e.getChannel());

	out.println("A new channel named "+e.getChannel().getChannelID()+
		    " was opened and is ready to be used.");
    }


/**
 * Aotomatic test execution area
 */

    private void startTest(StringTokenizer args) {

	if(args.countTokens()!=4) {
	    out.println("Invalid arguments number");
	    return;
	}

	try {
	    Channel c=findChannel(args.nextToken());
	    if(c==null) {
		out.println("Channel not found");
		return;
	    }
	
	    testModel=new TestEvent();
	    testModel.setChannel(c);
	    testModel.setDirection(new Direction(Direction.DOWN));
	    testModel.setSource(this);

	    testModel.source=new InetWithPort();
	    ((InetWithPort)testModel.source).host=InetAddress.getLocalHost();
	    ((InetWithPort)testModel.source).port=-1;
	    
	    testModel.dest=new InetWithPort();
	    ((InetWithPort)testModel.dest).host=
		InetAddress.getByName(args.nextToken());
	    ((InetWithPort)testModel.dest).port=
		Integer.parseInt(args.nextToken());

	    missingRetries=Integer.parseInt(args.nextToken());
	    
	    startTime=System.currentTimeMillis();

	    sendTestEvent();
	}
	catch(java.net.UnknownHostException e) {
	    out.println("One of the hosts is unknown. Test aborted.");
	}
    }

    private void sendTestEvent() {
	try {
	    TestEvent e=(TestEvent)testModel.cloneEvent();
	    
	    byte[] testNumber=String.valueOf(missingRetries).getBytes();

	    Message m=e.getMessage();
	    m.setByteArray(testNumber,0,testNumber.length);

	    MsgBuffer msgBuf=new MsgBuffer(null,0,1);
	    m.push(msgBuf);
	    msgBuf.data[msgBuf.off]=(byte)testNumber.length;
	    
	    e.init();
	    e.go();
	}
	catch(appia.AppiaEventException e) {
	    System.err.println("Unexpected event exception when sending test event");
	}
	catch(CloneNotSupportedException e) {
	    System.err.println("Appl: Clone Not Supported Exception");
	}
    }


    private void replyTest(TestEvent e) {

	try {
	    ReplyTestEvent reply=new ReplyTestEvent(e,this);

	    reply.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception while replying to test.");
	}
    }

    private void handleTestReply(ReplyTestEvent e) {
	/* Ensure the reply is correct */
	Message m=e.getMessage();
	MsgBuffer msgBuf=new MsgBuffer();

	/* Extract header length */
	msgBuf.len=1;
	m.pop(msgBuf);

	/* Extract header */
	msgBuf.len=msgBuf.data[msgBuf.off];

// 	out.println("Ponto da situacao. Comprimento da string: "+
// 		    msgBuf.len);

	m.pop(msgBuf);

// 	out.println("Bytes lidos: ");
// 	for(int i=msgBuf.off;i<msgBuf.off+msgBuf.len;i++)
// 	    out.print(msgBuf.data[i]);

	/* Make integer from header */
	String s=new String(msgBuf.data,msgBuf.off,msgBuf.len);
	int testVal=Integer.parseInt(s);

	/* Test the value */
	if(testVal!=missingRetries) {
	    out.println("Test failled. Expected number: "+
			missingRetries+" Received: "+testVal);
	    return;
	}
	
	if(--missingRetries!=0)
	    sendTestEvent();
	else
	    out.println("Test successfull. "+(System.currentTimeMillis()-startTime)+"ms");
    }

    private void handleUndelivered(FIFOUndelivered e) {
	try {
	    out.println("Endpoint "+((InetWithPort)e.who).host.getHostName()+
			" on port "+((InetWithPort)e.who).port+
			" apparently was not able to receive the "+
			"sent message(s). Give up");
	    e.go();
	}
	catch(AppiaEventException ex) {}
    }

    /**
     * Returns a pointer to the layer who created this session.
     */

    public Layer getLayer() {
	return layer;
    }
}
