package appia.protocols.fifo;

//////////////////////////////////////////////////////////////////////
//                                                                  //
// Appia: protocol development and composition framework            //
//                                                                  //
// Version: 1.0/J                                                   //
//                                                                  //
// Copyright, 2000, Universidade de Lisboa                          //
// All rights reserved                                              //
// See license.txt for further information                          //
//                                                                  //
// Class: FifoSession: Fifo ordering for multicast or unicast       //
//                                                                  //
// Author: Hugo Miranda, 05/2000                                    //
//                                                                  //
// Change Log:                                                      //
// Data, alteracao, razao/consequncias                             //
//                                                                  //
//////////////////////////////////////////////////////////////////////

import appia.*;
import appia.events.*;
import appia.events.channel.*;
import appia.message.*;
import appia.protocols.frag.MaxPDUSizeEvent;

import java.util.*;
import java.io.PrintStream;

/**
 * Class FifoSession provides the dynamic behaviour for the point to
 * point First In First Out Protocol. This protocol uses positive
 * acknowledgments to ensure correct ordering of messages. Out of
 * order messages are postponed and sent up when order is
 * reestablished. 
 *
 * For each end point address a java.util.Vector's element is kept
 * with a peer info object.
 *
 * Sequence number is started on zero. Since this is a test,
 * non-optimized class, no provision is made for reaching MAX_INT
 * 
 * Header format:
 * 4 bytes with sequence number, 4 bytes with acknowledgment. The fields 
 * are pushed with this order.


 * @author Hugo Miranda
 * @see    FifoLayer
 * @see    Session
 */

public class FifoSession extends Session {

    private final boolean debugOn=false;

    private Vector peers;

    /* Miliseconds until timeout of unacknowledge messages */
    private long   fifotimeout=1000;

    /* Milliseconds waited prior to send the acknowledgment */
    private long   piggybackdelay=500;

    /* Number of retries before giving up of sending messages to one peer */
    private int    nretries=5;

    private PrintStream debugOutput=null;

    /**
     * Mainly used for corresponding layer initialization
     */

    public FifoSession(Layer l) {
	super(l);

	peers=new Vector();
    }

    /**
     * This protocol does not attempt to bind any sessions on channels
     */

    public void boundSessions(Channel channel) {
    }

    /* Utils */

    /* Java has no "unsigned byte" so we can only use 7 of each 8 bits
       bytes. */

    private void seqToByte(MsgBuffer buf,int seq) {
	long factor=(long)Math.pow(2,21);
	int off=buf.off+3;

	for(int i=3;i>=0;i--,factor/=128,off--) {
	    buf.data[off]=(byte)(seq/factor);
	    seq-=buf.data[off]*factor;
	}
    }

    private int byteToSeq(MsgBuffer buf) {
	int acum=0,factor=1,off=buf.off;

	for(int i=0;i<=3;i++,factor*=128,off++) {
	    acum+=buf.data[off]*factor;
	}

	return acum;
    }

    private PeerInfo newPeer(Object who) {
	PeerInfo newpeer=new PeerInfo(who);
	peers.add(newpeer);
	return newpeer;
    }
 
    private PeerInfo findPeer(Object who) {

	PeerInfo p=null;
	java.util.Enumeration e=peers.elements();

	while(e.hasMoreElements() && p==null) {
	    PeerInfo t=(PeerInfo)e.nextElement();
	    if(((appia.protocols.udpsimple.InetWithPort)t.peer).equals((appia.protocols.udpsimple.InetWithPort)who)) p=t;
	}

	return p;
    }

    private void addNumbers(SendableEvent e,PeerInfo p) {

	/* Append header with sequence number */

	Message m=e.getMessage();

	/* Asks for 8 bytes and puts header in it*/

	MsgBuffer msgBuf=new MsgBuffer();
	msgBuf.len=8;
	m.push(msgBuf);
	seqToByte(msgBuf,p.nextOutgoing);
	msgBuf.off+=4;
	seqToByte(msgBuf,p.nextIncoming);
	
	/* Updates last acknowledgment sent to prevent 
	   duplicated acknowledgments */

	p.lastAckSent=p.nextIncoming;
    }

    private void processOutgoing(SendableEvent e) {
	
	PeerInfo p=findPeer(e.dest);

	/* New destiny. Allocates space for him */

	if(p==null) p=newPeer(e.dest);

	addNumbers(e,p);

	/* Duplicate the event and keep the copy to resend */

	WaitingMessage we=null;

	try {
	    we=p.addUnconfirmed((SendableEvent)e.cloneEvent(),p.nextOutgoing);
	}
	catch(CloneNotSupportedException ex) {
	    System.err.println("FIFO: Clone Not Supported Exception");
	}

	/* Prepares a timer. The timer will keep event and peer info
	   providing quick resend. */
	FIFOTimer ft=null;
	try {
	    ft=new FIFOTimer(e.getChannel(),new Direction(Direction.DOWN),
			     this,System.currentTimeMillis()+fifotimeout,
			     "FifoSession",we,p); 
	}
	catch(AppiaEventException ex) {
	    System.err.println("Channel does not accept timers "+
			       "(in FifoSession)");
	    switch(ex.type) {
	    case ex.NOTINITIALIZED :
		System.err.println("Event not initialized in message "+
				   "sending (FIFO Session)");
		break;
	    case ex.ATTRIBUTEMISSING :
		System.err.println("Missing attribute in message "+
				   "sending (FIFO Session)");
		break;
	    case ex.UNKNOWNQUALIFIER :
		System.err.println("Unknown qualifier (impossible) in "+
				   " message sending (FIFO Session)");
		break;
	    case ex.UNKNOWNSESSION :
		System.err.println("Unknown session in message "+
				   "sending (FIFO Session)");
		break;
	    case ex.UNWANTEDEVENT :
		System.err.println("Unwanted event in message "+
				   "sending (FIFO Session)");
		break;
	    case ex.CLOSEDCHANNEL :
		System.err.println("Channel closed in message "+
				   "sending (FIFO Session)");
		break;
	    }
	}
	catch(AppiaException ex) {
	    System.err.println("Unexpected appia generic exception in "+
			       "FifoSession");
	}

	try {
	    e.go();
	    ft.go();
//	    if(debugOn && debugOutput!=null)
//		debugOutput.println("FIFO: Event and FIFO Timer sent "+
//				    "to channel");
	} catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception in FifoSession");
	}

	p.incOutgoing();
    }

    /* CheckOrder returns true if the event was the next expected. The
     * goal is to dequeue only in this case else the event will be
     * added to existing queue. */ 

  private boolean checkOrder(PeerInfo p,SendableEvent e) {
    
    /* Extracting sequence number */

    MsgBuffer msgBuf=new MsgBuffer();
    msgBuf.len=8;
    e.getMessage().pop(msgBuf);

    int seqNumber=byteToSeq(msgBuf);

    msgBuf.off+=4;

    int confirmation=byteToSeq(msgBuf);

    /* Checks the piggybacked acknowledgment number */

    p.confirmedUntil(confirmation);
/*    if(debugOn && debugOutput!=null) 
	debugOutput.println("FIFO: Peer piggybacked confirmation until "+
			    confirmation);
*/
    /* Check if event is next expected */
    if(p.isNext(seqNumber)) {
      p.incIncoming();
      try {
	  e.go();
/*	  if(debugOn && debugOutput!=null)
	      debugOutput.println("FIFO: received event ("+seqNumber+
				  ") is the next expected. Forwarding it.");
*/      } catch(AppiaEventException ex) {
	  System.err.println("Unexpected event not initialized exception "+
			     "in FifoSession");
      }
      return true;
    }
    else {
	if(!p.isDuplicated(seqNumber)) {
	    p.enqueueIncoming(e,seqNumber);
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("FIFO: received event ("+seqNumber+
				    ") is not the next expected ("+
				    p.nextIncoming+"). enqueing it.");
*/	}
	else {
	    p.lastAckSent=Math.min(seqNumber,p.lastAckSent);
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("FIFO: Duplicated message ("+
				    seqNumber+") received");
*/	}
    }
    return false;
  }
    
    private void dequeue(PeerInfo p) {
	SendableEvent e;

	while((e=p.dequeueNextIncoming())!=null) {
	    p.incIncoming();
	    try {
		e.go();
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("FIFO: Dequeing next event");
*/	    } catch(AppiaEventException ex) {
		System.err.println("Unexpected event in FifoSession");
	    }
	}
    }

    /* Cummulative acknowledgment send */

    private void sendAck(SendAckTimer e) {
	AckEvent ack=null;
	PeerInfo p=e.peer;

	try {
	    ack=new AckEvent(e,this);
	    Message m=new Message();
	    MsgBuffer msgBuf=new MsgBuffer();
	    msgBuf.len=4;
	    m.push(msgBuf);
	    seqToByte(msgBuf,p.nextIncoming);
	    ack.setMessage(m);
	    ack.go();

	    p.lastAckSent=p.nextIncoming;

/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("FIFO: ack sent");
*/	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception in FifoSession");
	}
    }
	
    private void handleSendAckTimer(SendAckTimer e) {
	/* Timer should be discarded if:
	   - it is not an HAPPENED event
	   - the acknowledgments sent are later than the requested
	*/

/*	if(debugOn && debugOutput!=null)
	    debugOutput.print("FIFO: Received request for sending "+
			      "acknowledgment. ");
*/	if(e.getQualifier().isNotify() && 
	   e.peer.lastAckSent<e.holdingSeqNumber) {
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Ack will be sent");
*/	    sendAck(e);
	}
	else {
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Discarded");
*/	}
    }

    private void processIncoming(SendableEvent e)
    {
	PeerInfo p=findPeer(e.source);
	
	if(p==null)  p=newPeer(e.source);

	if(checkOrder(p,e))
	    dequeue(p);

	/* Prepares a timer to send an acknowledgment case there's no
	   reply from upper layers */
	try {
	    SendAckTimer sat=new 
		SendAckTimer(System.currentTimeMillis()+piggybackdelay,
			     e.getChannel(),this,p,p.nextIncoming,e.dest);
	    sat.go();
/*	    if(debugOn && debugOutput!=null) 
		debugOutput.println("FIFO: The request of a timer for "+
				    "unpiggibacked acknowledgments "+
				    "was made.");
*/	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception when creating "+
			       "SendAckTimer");
	}
	catch(AppiaException ex) {
	    System.err.println("Unexpected exception when creating "+
			       "SendAckTimer");
	}
    }
	    
    /**
     * Main Event handler function. Accepts all incoming events and
     * dispatches them to the appropriate functions
     * @ params e The incoming event
     * @ see Session
     */

    public void handle(Event e) {
	if(e instanceof AckEvent)
	    handleAck((AckEvent)e);
	else if(e instanceof ChannelInit)
	    handleChannelInit((ChannelInit)e);
	else if(e instanceof FIFOTimer)
	    handleFIFOTimer((FIFOTimer)e);
	else if(e instanceof SendAckTimer)
	    handleSendAckTimer((SendAckTimer)e);
	else if(e instanceof SendableEvent)
	    handleSendable((SendableEvent)e);
	else if(e instanceof Debug)
	    handleDebug((Debug)e);
	else if(e instanceof MaxPDUSizeEvent)
	    handlePDUSize((MaxPDUSizeEvent)e);
	/* Debug */
	else {
	    /* Unexpected event arrived */
	    try {
		e.go();
	    }
	    catch(AppiaEventException ex) {}
	}
	
/*	if(debugOn && debugOutput!=null)
	    debugOutput.println("FIFO: event handling concluded");
*/    }

    private void handlePDUSize(MaxPDUSizeEvent e) {
	try {
	    /* Subtract to the actual PDUSize the biggest possible header */
	    e.pduSize-=8;
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception when forwarding "+
			       "MaxPDUSize event in FIFO");
	}
    }

    private void handleDebug(Debug e) {

	EventQualifier q=e.getQualifier();

	if(q.isOn()) {
	    debugOutput=new PrintStream(e.getOutput());
	    debugOutput.println("FIFO: Debugging started");
	}
	else if(q.isOff()) {
	    debugOutput=null;
	}
	else if(q.isNotify()) {
	    printState(new PrintStream(e.getOutput()));
	}

	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	}
    }

    private void printState(PrintStream out) {
	out.println("FIFO Session state dumping:");

	out.println("Retransmission delay: "+fifotimeout+"ms");

	out.println("Current number of peers: "+peers.size());

	Enumeration en=peers.elements();
	int count=0;
	while(en.hasMoreElements()) {
	    PeerInfo p=(PeerInfo)en.nextElement();
	    out.println("Host "+count+":");
	    out.println("  Next sequence number to be sent: " + 
			p.nextOutgoing);
	    out.println("  Next sequence number expected: "+p.nextIncoming);
	    out.println("  First message still waiting for "+
			"acknowledgment: "+p.firstUnconfirmed);
	    out.println("  There are "+p.getWaitingQueueSize()+ 
			" messages waiting to be delivered by the "+
			"proper order.");
	    count++;
	}
	out.println("Debug output is currently "+ 
		    (debugOutput==null ? "off" : "on"));
    }

	
    /**
     * Acknowledgment event received. Will be used to delete all
     * pending messages waiting for confirmation. This version does
     * not attempt to cancel timeout events
    */

    private void handleAck(AckEvent e) {
	if(debugOn && debugOutput!=null)
	    debugOutput.println("FIFO: Ack event received");
	/* Sanity checks */
	if(e.getDirection().direction!=Direction.UP)
	    /* Oops, this is not our acknowledge */
	    handleSendable((SendableEvent)e);
	else {
	    PeerInfo p=findPeer(e.source);
	    if(p!=null) { /* False is only expected when the peer has failed.
			     Ignored */
		MsgBuffer msgBuf=new MsgBuffer();
		msgBuf.len=4;
		e.getMessage().pop(msgBuf);
		int confirmation=byteToSeq(msgBuf);
		p.confirmedUntil(confirmation);
		if(debugOn && debugOutput!=null)
		    debugOutput.println("FIFO: Ack received until sequence "+
					"number "+confirmation);
	    }
	}
    }
	    
    /**
     * As long as an event is sendable,
     * events are all treated the same way: a sequence number is
     * pushed on the header.
     *
     * @param e The SendableEvent received.
     */

    private void handleSendable(SendableEvent e) {
/*	if(debugOn && debugOutput!=null)
	    debugOutput.println("FIFO: sendable event received");
*/
 	switch(e.getDirection().direction) {
 	case Direction.UP :
 	    processIncoming(e);
	    break;
	case Direction.DOWN :
	    processOutgoing(e);
	    break;
	default:
	    /* Invalid direction */
	}
    }

    /**
     * This event is only handled in order to inform the layer that
     * this session is being used on a channel. Since this is an
     * channel unaware protocol, information collected about channels
     * is no further handled.
     *
     * @param e The ChannelInit event
     */

    private void handleChannelInit(ChannelInit e) {
/*	if(debugOn && debugOutput!=null)
	    debugOutput.println("FIFO: channel init");
*/	// Next line commented to avoid Class Cast Exception with Ext version
	// of layer
	//((FifoLayer)layer).linkChannelWithSession(e.getChannel(),this);

	try {
	    e.go();
	} catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception in FifoSession");
	}
    }

    /**
     * Handling of timeout events. Receiving a timeout event does not
     * necessarily means that the message wasn't
     * acknowledge. Timer cancels can be delayed on the channel
     */

    private void handleFIFOTimer(FIFOTimer timeout) {
	if(debugOn && debugOutput!=null)
 	    debugOutput.println("FIFO: Fifo Timer received");

	/* Check if the event is moving on the right direction and if
	   the message is still awaiting to be acknowledged */
	if(timeout.getDirection().direction==Direction.UP && 
	   !timeout.getPeerInfo().failed && 
	   !timeout.getPeerInfo().isAcknowledged(timeout.getWait())) {

	    if(timeout.retry<nretries)
		resend(timeout);
	    else
		giveup(timeout.getChannel(),timeout.getPeerInfo());
	}
	else
	     if(debugOn && debugOutput!=null)
		 debugOutput.println("FIFO: Unnecessary timeout. Message "+
				     " already acknowledged or peer failed");
	
    }

    private void resend(FIFOTimer timeout) {
	/* Resend message. FIFOTimeout what's attribute
	   contains an WaitingMessage object which in turn contains
	   a copy of the unacknowledged message. Make another
	   copy and tell it to go again. */
	try {
	    SendableEvent unack=
		(SendableEvent)timeout.getWait().e.cloneEvent();
	    /* The copy is not a "real" clone. Generator has changed */
	    unack.setSource(this);
	    unack.init();
	    unack.go();
	    if(debugOn && debugOutput!=null)
		debugOutput.println("FIFO: Unacknowledged message resent");

	/* Also perform a duplication of the timeout to request
	   a new one */
	    FIFOTimer newtimeout=timeout;
	    newtimeout.retry++;
	    newtimeout.setTimeout(System.currentTimeMillis() +
				  fifotimeout);
	    newtimeout.setDirection(new Direction(Direction.DOWN));
	    newtimeout.setSource(this);
	    newtimeout.getQualifier().set(EventQualifier.ON);
	    newtimeout.init();
	    newtimeout.go();
	    if(debugOn && debugOutput!=null)
		debugOutput.println("FIFO: New timeout set");

	}
	catch(CloneNotSupportedException ex) {
	    System.err.println("FIFO: Clone not supported exception");
	}
	catch(AppiaEventException ex) {
	    switch(ex.type) {
	    case AppiaEventException.UNWANTEDEVENT :
		System.err.println("Unwanted event exception catched "+
				   "in FifoSession");
		break;
	    case AppiaEventException.UNKNOWNSESSION :
		System.err.println("Unknown session exception catched "+
				   "in FifoSession");
		break;
	    case AppiaEventException.ATTRIBUTEMISSING :
		System.err.println("Missing attribute exception catched "+
				   "in FifoSession");
		break;
	    case AppiaEventException.NOTINITIALIZED :
		System.err.println("Impossible exception event not "+
				   "initialized in FifoSession");
		break;
	    }
	}
	catch(AppiaException ex) {
	    System.err.println("Negative time in event definition");
	}
    }

    /* Removes a PeerInfo from the list of peers. Gives up of
       sending messages to him
    */
    private void giveup(Channel c,PeerInfo p) {
	/* If the peer exists in the vector, removes it and
	 notifies sessions above with an event */
	if(peers.remove(p)) {
	    p.failed=true;
	    try {
		FIFOUndelivered e=new FIFOUndelivered(c,this,p.peer);
		e.go();
	    }
	    catch(AppiaEventException ex) {
		switch(ex.type) {
		case AppiaEventException.UNWANTEDEVENT :
		    /* This one is ok. No one cares about it */
		    break;
		case AppiaEventException.UNKNOWNSESSION :
		    System.err.println("Unknown session exception catched "+
				       "in FifoSession");
		    break;
		case AppiaEventException.ATTRIBUTEMISSING :
		    System.err.println("Missing attribute exception catched "+
				       "in FifoSession");
		    break;
		case AppiaEventException.NOTINITIALIZED :
		    System.err.println("Impossible exception event not "+
				       "initialized in FifoSession");
		    break;
		}
	    }
	}
    }

    /**
     * Returns a pointer to the layer who created this session.
     */
    
    public Layer getLayer() {
	return layer;
    }
}
