package appia.protocols.prdAppl;

//////////////////////////////////////////////////////////////////////
//                                                                  //
// Appia: protocol development and composition framework            //
//                                                                  //
// Version: 1.0/J                                                   //
//                                                                  //
// Copyright, 2000, Universidade de Lisboa                          //
// All rights reserved                                              //
// See license.txt for further information                          //
//                                                                  //
// Class: ApplicationSession: User interface with Appia             //
//                                                                  //
// Author: Hugo Miranda, 05/2000                                    //
//                                                                  //
// Change Log:                                                      //
// Data, alteracao, razao/consequncias                             //
//                                                                  //
//////////////////////////////////////////////////////////////////////

import appia.*;
import appia.events.*;
import appia.events.channel.*;
import appia.message.*;
import appia.protocols.udpsimple.RegisterSocketEvent;
import appia.protocols.udpsimple.InetWithPort;
import appia.protocols.group.*;
import appia.protocols.group.events.*;

import java.io.*;
import java.util.*;
import java.net.InetAddress;

/**
 * Class ApplicationSession provides the dynamic behaviour for the 
 * simple user interface with Appia. This interface accepts a predefined
 * set of commands, forwarding them to Appia. It is mainly used for testing
 * @author Hugo Miranda
 * @see    ApplicationLayer
 * @see    Session
 */

public class PrdApplSession extends Session {

    private final int INVALIDCOMMAND=-1;
    private final int CASTMESSAGE=1;
    private final int HELP=2;
    private final int DEBUGMESSAGE=3;
    private final int SENDMESSAGE=4;

    private final int CAST=-1;

    /* User IO */
    private PrintStream out=System.out;

    /* Appia */

    public  Channel       channel=null;
    private int           myPort;
    private PrdApplReader reader;

    /* Group */
    ViewState    vs;
    int          rank;
    boolean      isBlocked=true;

    /**
     * Mainly used for corresponding layer initialization
     */

    public PrdApplSession(PrdApplLayer l) {
	super(l);

	reader=new PrdApplReader(this);
	reader.start();
    }

    public void init(int myPort,int rank,String[] view,int off) 
	throws AppiaGroupException, java.net.UnknownHostException {

	this.myPort=myPort;
	this.rank=rank;
	makeView(rank,view,off);
    }

    private int parseCommand(String command) {
	if(command.equals("cast"))
	    return CASTMESSAGE;
	if(command.equals("debug"))
	    return DEBUGMESSAGE;
	if(command.equals("help"))
	    return HELP;
	if(command.equals("send"))
	    return SENDMESSAGE;
	return INVALIDCOMMAND;
    }

    private void printHelp() {
	out.println("help\tImprime esta mensagem");
	out.println("cast r i m\t Envia mensagem m r vezes "+
		    "a intervalos de i decimas de segundo");
	out.println("send i r d m\t Envia a mensagem m r vezes para o membro de indice i, a intervalos de d decimos de segundo");
	out.println("debug {start|stop|now} [ficheiroOutput]");
    }


    private void sendDebug(StringTokenizer args) {
	if(args.countTokens()!=1 && args.countTokens()!=2) {
	    out.println("Invalid arguments number. Usage:");
	    out.println("debug one of this options: start, "+
			"stop, now and an optional file name");
	    return;
	}
	
	try {
	    EventQualifier eq=new EventQualifier();

	    String action=args.nextToken();
	    
	    if(action.equals("start"))
		eq.set(eq.ON);
	    else if(action.equals("stop"))
		eq.set(eq.OFF);
	    else if(action.equals("now"))
		eq.set(eq.NOTIFY);
	    else {
		out.println("Invalid action. Use one of start, stop and now");
		return;
	    }
	    
	    OutputStream debugOut=out;

	    if(args.hasMoreTokens()) {
		try {
		    debugOut=new FileOutputStream(args.nextToken(),false);
		}
		catch(FileNotFoundException ex) {
		    out.println("File could not be oppened for output");
		    return;
		}
	    }

	    appia.events.channel.Debug e=
		new appia.events.channel.Debug(debugOut);
	    e.setChannel(channel);
	    e.setDirection(new Direction(Direction.DOWN));
	    e.setSource(this);
	    e.setQualifier(eq);
	    
	    e.init();	    
	    e.go();
	}
	catch(appia.AppiaEventException ex) {
	    out.println("Unexpected exception when sending debug event");
	}
    }
	
    private void castMessage(StringTokenizer args,String commandLine) {
	if(args.countTokens()<3) {
	    out.println("Numero de argumentos invalido para cast");
	    return;
	}

	int period=0,resends=0;

	try {
	    resends=Integer.parseInt(args.nextToken());
	    period=Integer.parseInt(args.nextToken())*100;
	}
	catch(NumberFormatException ex) {
	    out.println("O numero de repeticoes e o intervalo tem que ser "+
			"numeros inteiros. Digite \"ajuda\" para obter "+
			"mais informacao");
	    return;
	}

	String mensagem=new String();

	while(args.hasMoreElements()) {
	    mensagem+=args.nextToken()+" ";
	}

	try {
	    PrdApplTimer msgTimer=new PrdApplTimer(channel,this,mensagem,
						   resends,period,CAST);
	    msgTimer.go();
	}
	catch(AppiaEventException ex) {
	    out.println("Houve problema com o envio das mensagens.");}
	catch(AppiaException ex) {
	    out.println("O intervalo tem que ser >=0");
	}
    }

    private void sendMessage(StringTokenizer args,String commandLine) {
	if(args.countTokens()<4) {
	    out.println("Numero de argumentos invalido para send");
	    return;
	}

	int period=0,resends=0,dest=CAST;

	try {
	    dest=Integer.parseInt(args.nextToken());
	    resends=Integer.parseInt(args.nextToken());
	    period=Integer.parseInt(args.nextToken())*100;
	}
	catch(NumberFormatException ex) {
	    out.println("O indice do destinatario, numero de repeticoes e o intervalo tem que ser "+
			"numeros inteiros. Digite \"ajuda\" para obter "+
			"mais informacao");
	    return;
	}

	String mensagem=new String();

	while(args.hasMoreElements()) {
	    mensagem+=args.nextToken()+" ";
	}

	try {
	    PrdApplTimer msgTimer=new PrdApplTimer(channel,this,mensagem,
						   resends,period,dest);
	    msgTimer.go();
	}
	catch(AppiaEventException ex) {
	    out.println("Houve problema com o envio das mensagens.");}
	catch(AppiaException ex) {
	    out.println("O intervalo tem que ser >=0");
	}
    }


    /**
     * Parse and send user messages
     */

    private void parseUser() {
	synchronized(reader) {
	    if(reader.sComLine!=null) {
		StringTokenizer comLine=new StringTokenizer(reader.sComLine);
		
		if(comLine.hasMoreTokens()) {
		    String command=comLine.nextToken();
		    
		    switch(parseCommand(command)) {
		    case HELP :
			printHelp();
			break;
		    case CASTMESSAGE :
			castMessage(comLine,reader.sComLine);
			break;
		    case DEBUGMESSAGE :
			sendDebug(comLine);
			break;
		    case SENDMESSAGE:
			sendMessage(comLine,reader.sComLine);
			break;
		    default :
			System.out.println("Invalid command");
		    }
		}
		reader.sComLine=null;
		reader.notify();
	    }
	}
    }

    /**
     * Main Event handler function. Accepts all incoming events and
     * dispatches them to the appropriate functions
     * @ params e The incoming event
     * @ see Session
     */

    public void handle(Event e) {
	if(e instanceof ChannelInit) 
	    handleChannelInit((ChannelInit)e);
	else if(e instanceof PrdApplCastEvent)
	    receiveData((PrdApplCastEvent)e);
	else if(e instanceof PrdApplSendEvent)
	    receiveData((PrdApplSendEvent)e);
	else if(e instanceof View)
	    handleNewView((View)e);
	else if(e instanceof BlockOk)
	    handleBlock((BlockOk)e);
	else if(e instanceof PrdApplTimer)
	    handleTimer((PrdApplTimer)e);

	if(!isBlocked)
	    parseUser();
    }
    
    private void handleNewView(View e) {
	vs=e.vs;
	rank=e.ls.my_rank;
	isBlocked=false;

	out.println("Instalada nova vista:");
	out.println("Lista de membros (por IP:porto):");
	for(int i=0;i<vs.addresses.length;i++)
	    out.println("{"+vs.addresses[i].host.getHostAddress()+
			":"+vs.addresses[i].port+"} ");
	out.println((e.ls.am_coord ? "Sou" : "Nao sou") + " o coordenador "+
		    "do grupo");
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    out.println("Excepcao ao mandar o evento de nova vista");
	}
    }

    private void handleBlock(BlockOk e) {
	parseUser();
	out.println("O grupo foi bloqueado. Temporariamente nao e possivel "+
		    "enviar mensagens.");
	isBlocked=true;
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    out.println("Excepcao ao fazer forward do evento block ok");
	}
    }

    private void handleTimer(PrdApplTimer e) {

	if(isBlocked) {
	    try {
		e.setSource(this);
		e.getDirection().invert();
		e.getQualifier().set(EventQualifier.ON);
		e.setTimeout(System.currentTimeMillis()+e.period);
		e.init();
		e.go();
	    }
	    catch(AppiaEventException ex) {
		out.println("Excepcao ao mandar o evento timer com o grupo bloqueado");
	    }
	    catch(AppiaException ex) {
		out.println("Excepcaoao mandar o evento timer com o grupo bloqueado");
	    }
	    return;
	}
	try {
	    GroupSendableEvent msgEvent;

	    if (e.dest == CAST)
		msgEvent=new PrdApplCastEvent(channel,this,vs.group,vs.id);
	    else {
		int[] dests={e.dest};
		msgEvent=new PrdApplSendEvent(channel,this,vs.group,vs.id,dests);
	    }
	    
	    String msg=new String(e.msg+e.thisResend);
	    
	    byte[] data=new byte[msg.length()+4];
	    lengthToByte(data,msg.length());
	    System.arraycopy(msg.getBytes(),0,data,4,msg.length());
	
	    msgEvent.getMessage().setByteArray(data,0,data.length);
	    msgEvent.go();
	    //	    out.println("Enviada mensagem: "+new String(data));
	    if(e.hasMore()) {
		e.prepareNext();
		e.go();
	    }
	}
	catch(appia.AppiaEventException ex) {
	    System.err.println("Unexpected exception in Application Session");
	    switch(ex.type) {
	    case appia.AppiaEventException.NOTINITIALIZED :
		System.err.println("Event not initialized in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.ATTRIBUTEMISSING :
		System.err.println("Missing attribute in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.UNKNOWNQUALIFIER :
		System.err.println("Unknown qualifier (impossible) in "+
				   " message sending (Application Session)");
		break;
	    case appia.AppiaEventException.UNKNOWNSESSION :
		System.err.println("Unknown session in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.UNWANTEDEVENT :
		System.err.println("Unwanted event in message "+
				   "sending (Application Session)");
		break;
	    case appia.AppiaEventException.CLOSEDCHANNEL :
		System.err.println("Channel closed in message "+
				   "sending (Application Session)");
		break;
	    }
	}
	catch(AppiaException ex) {
	    out.println("Excepcao ao mandar um timer para a proxima mensagem");
	}
    }

    private void lengthToByte(byte[] header,int length) {
	long factor=(long)Math.pow(2,24);

	for(int i=3;i>=0;i--,factor/=256) {
	    header[i]=(byte)(length/factor);
	    length-=header[i]*factor;
	}
    }

    private int byteToLength(MsgBuffer buf) {
	int acum=0;
	long factor=1;
	int  off=buf.off;

	for(int i=0;i<=3;i++,factor*=256,off++) {
	    acum+=buf.data[off]*factor;
	}

	return acum;
    }

    private void receiveData(SendableEvent e) {

	/* Echoes received messages to the user */
	if(e.getDirection().direction==Direction.UP) {
	    Message m=e.getMessage();
	    MsgBuffer msgBuf=new MsgBuffer();
	    msgBuf.len=4;
	    m.pop(msgBuf);
	    int l=byteToLength(msgBuf);

	    if(e instanceof PrdApplCastEvent)
		out.println("Mensagem de multicast, recebida de "+((Endpt)e.source).id);
	    else
		out.println("Mensagem ponto-a-ponto, recebida de "+((Endpt)e.source).id);

	    while(l>0) {
		msgBuf.len=l;
		m.pop(msgBuf);
		out.print(new String(msgBuf.data,msgBuf.off,msgBuf.len));
		l-=msgBuf.len;
	    }
	    System.out.println();
	}
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception in Application "+
			       "session");
	}
    }
	    

    private void handleChannelInit(ChannelInit e) {

	/* Forwards channel init event. New events must follow this
	   one */
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception in Application "+
			       "session");
	}

	channel=e.getChannel();

	/* Informs layers bellow of the port
	   where messages will be received and sent */
       
	try {
	    RegisterSocketEvent rse=new
		RegisterSocketEvent(channel,new Direction(Direction.DOWN),
				    this,myPort);
	    rse.go();
	}
	catch(appia.AppiaEventException ex) {
	    switch(ex.type) {
	    case appia.AppiaEventException.UNWANTEDEVENT :
		System.err.println("The QoS definition doesn't satisfy the "+
				   "application session needs. "+
				   "RegisterSocketEvent, received by "+
				   "UdpSimpleSession is not being acepted");
		break;
	    default :
		System.err.println("Unexpected exception in "+
				   "Application session");
		break;
	    }
	}

	/* Lanca o evento de inicializacao do grupo */
	//	if(rank==0) {
	    try {
		GroupInit gi=new GroupInit(vs,vs.view[rank],null,
					   channel,
					   new Direction(Direction.DOWN),this);
		gi.go();
	    }
	    catch(AppiaEventException ex) {
		System.err.println("EventException no lancamento do "+
				   "GroupInit");
	    }
	    catch(NullPointerException ex) {
		System.err.println("NullPointerException no lancamento "+
				   "do GroupInit");
	    }
	    catch(AppiaGroupException ex) {
		System.err.println("AppiaGroupException no lancamento do "+
				   "GroupInit");
	    }
	    //	}

	out.println("Aberto canal com o nome "+
		    e.getChannel().getChannelID());
    }

    private void makeView(int rk,String[] view,int off) 
	throws AppiaGroupException, java.net.UnknownHostException {
	
	Endpt coord=new Endpt("0");
	InetWithPort[] enderecos=makeEnderecos(view,off);
	Endpt[] vista=makeVista(view.length-off);

	vs=new ViewState("versao1",new Group("grupoID"),
			 new ViewID(0,coord),new ViewID[0],
			 vista,enderecos);
    }

    private InetWithPort[] makeEnderecos(String[] view,int off) 
	throws IndexOutOfBoundsException, java.net.UnknownHostException {

	int nelem=view.length-off;
	InetWithPort[] ret=new InetWithPort[nelem];
	for(int i=0;i<nelem;i++) {
	    try {
		ret[i]=new InetWithPort();
		String host=
		    view[i+off].substring(0,view[i+off].indexOf(':'));
		ret[i].host=InetAddress.getByName(host);
		ret[i].port=
		    Integer.parseInt(view[i+off].substring(view[i+off].indexOf(':')+1));
	    }
	    catch(StringIndexOutOfBoundsException ex) {
		System.err.println("Formato host:porto nao foi respeitado "+
				   "no elemento "+view[i+off]);
		System.exit(1);
	    }
	}
	return ret;
    }

    private Endpt[] makeVista(int quantos) {
	Endpt[] ret=new Endpt[quantos];
	for(int i=0;i<quantos;i++) {
	    ret[i]=new Endpt(String.valueOf(i));
	}
	return ret;
    }

    /**
     * Returns a pointer to the layer who created this session.
     */

    public Layer getLayer() {
	return layer;
    }
}
