package appia.protocols.udpsimple;

//////////////////////////////////////////////////////////////////////
//                                                                  //
// Appia: protocol development and composition framework            //
//                                                                  //
// Version: 1.0/J                                                   //
//                                                                  //
// Copyright, 2000, Universidade de Lisboa                          //
// All rights reserved                                              //
// See license.txt for further information                          //
//                                                                  //
// Class: UdpSimpleSession: Unreliable send/receive using UDP/IP    //
//                                                                  //
// Author: Hugo Miranda, 05/2000                                    //
//                                                                  //
// Change Log:                                                      //
// Data, alteracao, razao/consequncias                             //
//                                                                  //
//////////////////////////////////////////////////////////////////////

import appia.*;
import appia.events.*;
import appia.events.channel.*;
import appia.protocols.application.*;
import appia.protocols.frag.MaxPDUSizeEvent;

import java.net.*;
import java.util.*;
import java.io.*;

/**
 * Class UdpSimpleSession is the Session subclassing for UdpSimple
 * protocol. Concurrently to this session, a reader class, running on
 * another thread listens the socket and esqueues
 * datagrams. Synchronization with the Appia is made using events:
 * whenever an event is being processed, the session checks the thread
 * and dequeues readed datagrams. Sended messages are organized in the
 * following way: event class name, channel id, sender, dest,
 * msg. All fields except msg are prefixed by a byte with their
 * length. dest and source are expected to be InetAddress objects on
 * SendableEvents and are coded in a string because InetAddress
 * doesn't have a constructor from a byte[].
 *
 * @see appia.Session
 * @author Hugo Miranda
 */

public class UdpSimpleSession extends Session {

    private final boolean debugOn=true;

    private DatagramSocket mySocket=null;
    protected Vector channels=null,closingChannels=null;
    private UdpSimpleReader sReader=null;
    private int port;

    private boolean channelsToClose=false;

    private PrintStream debugOutput=null;

    /* 20 to test fragmentation layer */
    private final int pduSize=60000;

    public UdpSimpleSession(Layer l) {
	super(l);
	
	channels=new Vector();
	closingChannels=new Vector();

	if(debugOn && debugOutput!=null)
	    debugOutput.println("UDP: New udpsimple session");
    }

    public Layer getLayer() {
	return layer;
    }

    public void boundSessions(Channel c) {}

    /**
     * The event handler function. Tests event types and dispatches
     * them to the appropriate handler.
     * @param e The event
     * @see Session.handle
     */

    public void handle(Event e) {
	if(e instanceof RegisterSocketEvent)
	    handleRegisterSocket((RegisterSocketEvent) e);
	else if(e instanceof SendableEvent)
	    handleSendable((SendableEvent) e);
	else if(e instanceof ChannelInit)
	    handleChannelInit((ChannelInit) e);
	else if(e instanceof ChannelClose)
	    handleChannelClose((ChannelClose) e);
	else if(e instanceof AsyncEvent)
	    handleAsync((AsyncEvent)e);
	else if(e instanceof Debug)
	    handleDebug((Debug)e);
	else if(e instanceof MaxPDUSizeEvent)
	    handlePDUSize((MaxPDUSizeEvent)e);
	else {
	    /*Unexpected event received in UdpSimpleSession */
	    try {
		e.go();
	    }
	    catch(AppiaEventException ex) {}
	}
    }

    private void handlePDUSize(MaxPDUSizeEvent e) {
	try {
	    /* Set the maximum PDU size acceptable. Defined by
	       a constant in the attributes section of the class */
	    e.pduSize=pduSize;
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception when forwarding "+
			       "MaxPDUSize event in UDPSimple");
	}
    }

    private void handleDebug(Debug e) {

	EventQualifier q=e.getQualifier();

	if(q.isOn()) {
	    debugOutput=new PrintStream(e.getOutput());
	    sReader.debugOutput=debugOutput;	    
	    debugOutput.println("UDP: Debugging started");
	}
	else if(q.isOff()) {
	    debugOutput=null;
	    sReader.debugOutput=null;
	}
	else if(q.isNotify()) {
	    printState(new PrintStream(e.getOutput()));
	}

	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	}
    }

    private void printState(PrintStream out) {
	out.println("UdpSimpleSession state dumping:");
	out.println("Local UDP port: "+mySocket.getLocalPort());

	int nChannels=channels.size();
	out.println("Currently connected channels: "+nChannels);
	
	java.util.Enumeration en=channels.elements();
	while(en.hasMoreElements()) {
	    Channel c=(Channel)en.nextElement();
	    out.println("Channel name: "+c.getChannelID()+" QoS: "+
			c.getQoS().getQoSID());
	}
	
	out.println("Debug output is currently "+ 
		    (debugOutput==null ? "off" : "on"));
    }
    

    /**
     * Receives a register socket request. Only one of this events is
     * expected to be received by channel. 
     *
     * @param e The socket description and the ID which will be used
     * to identify it. 
     */

    private void handleRegisterSocket(RegisterSocketEvent e) {

	/* if the socket is already binded then something is
	 * wrong. Keep existing information.
	 */
	if(debugOn && debugOutput!=null) 
	    debugOutput.println("UDP: register socket event");

	if(mySocket!=null)
	    return;
	
	port=e.port;

	/* Open the specified socket (if possible) */

	try {
	    mySocket=new DatagramSocket(port);
	}
	catch(SocketException se) {
	    /* If a socket exception occurs, the socket cannot be
	     * opened. Create an error event to notify the requester
	     */
	    return;
	}

	/* The socket is binded. Launch reader and request a heartbeat.
	 */

	sReader=new UdpSimpleReader(this,mySocket);
	sReader.start();

	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Received event not initialized in "+
			       "UdpSimpleSession");
	}
	//	checkMessages(e.getChannel());
    }

    /** Receives sendable events from sessions above and puts them on
     * the socket. 
     *
     * @param e The event to be sent
     */

    private void handleSendable(SendableEvent e) {

	if(debugOn && debugOutput!=null)
	    debugOutput.println("UDP: Sendable Event");

	if(e.getDirection().direction==Direction.DOWN) {

	    formatAndSend(e);
	
	    /* Now that the packet is sent, event follows his way */
	    try {
		e.go();
	    }
	    catch(AppiaEventException ex) {
		System.err.println("Event not initialized but tried to be "+
				   "sent in UdpSimpleSession");
	    }
	}

	//	checkMessages(e.getChannel());
    }
    
    /**
     * Handle of ChannelInit event. Keep tracks of which channels this
     * session is currently serving to learn where to route the
     * incoming messages. Notifies his layer.
     *
     * @param e The ChannelInit event
     */

    private void handleChannelInit(ChannelInit e) {

	if(debugOn && debugOutput!=null)
	    debugOutput.println("UDP: Channel Init");

	// Next line commented to avoid Class Cast Exception with Ext version
	// of layer
	//(UdpSimpleLayer)layer).linkChannelWithSession(e.getChannel(),this);
	channels.add(e.getChannel());

	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Event not initialized exception in "+
			       "UdpSimpleSession");
	}
	
	if(debugOn && debugOutput!=null)
	    debugOutput.println("UDP: channel init handling concluded");
    }
    

    /* Starts to remove the channel from the vector. The channel can not 
       be immediatily removed because a pending message may exist on the
       reader. This would deadlock the reader. An async is requested instead.
       When we now that the reader is locked we remove the channel.
    */

    private void handleChannelClose(ChannelClose e) {
	
	if(debugOn && debugOutput!=null)
	    debugOutput.println("UDP: Channel "+e.getChannel().getChannelID()+
				" closed");
	closingChannels.add(e);
	channelsToClose=true;
    }

    /* For each channel awaiting to be closed:
       - remove the CloseChannel event from the vector
       - remove the Channel from the channel's vector
       - forward the event
    */

    private void removeChannels() {
        while(!closingChannels.isEmpty()) {
	    ChannelClose e=(ChannelClose)closingChannels.remove(0);
	    channels.remove(e.getChannel());
	    try {
		e.go();
	    }
	    catch(AppiaEventException ex) {
		System.err.println("Unexpected exception when "+
				   "forwarding ChannelClose event");
	    }
	}
	channelsToClose=false;
    }

    private void handleAsync(AsyncEvent e) {

	checkMessages(e.getChannel());
	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected exception in UdpSimple Session");
	}
    }

    /** Handle of messages from the network.
     */
    
    private void checkMessages(Channel forChannel) {
	synchronized(sReader) {
	    /* Only inserts events when the channel that gaves the thread
	       to this session is the one the event is waiting
	    */
	    if(sReader.waitingEvent!=null && 
	       sReader.waitingEvent.getChannel()==forChannel) {
		insertOnChannel(sReader.waitingEvent);
		/* This is the only moment when is safe to remove a
		   channel from the channels vector
		*/
		if(channelsToClose)
		    removeChannels();
		sReader.waitingEvent=null;
		sReader.notify();
	    }
	}
    }

    /** Event serialization and send to socket
     */
    private void formatAndSend(SendableEvent e) {
	ByteArrayOutputStream data=new ByteArrayOutputStream();

	/* Event Class name */

	String eventType=e.getClass().getName();
	byte[] bytes=eventType.getBytes();
	data.write(bytes.length);
	data.write(bytes,0,bytes.length);

	/* Channel ID */
	
	bytes=e.getChannel().getChannelID().getBytes();
	data.write(bytes.length);
	data.write(bytes,0,bytes.length);

	/* Sender */

	bytes=((InetWithPort)e.source).host.getHostAddress().getBytes();
	data.write(bytes.length);
	data.write(bytes,0,bytes.length);

	/* Dest */

	bytes=((InetWithPort)e.dest).host.getHostAddress().getBytes();
	data.write(bytes.length);
	data.write(bytes,0,bytes.length);

	/* Message */

	bytes=e.getMessage().toByteArray();
	data.write(bytes,0,bytes.length);

	/* Create the packet and send it */

	bytes=data.toByteArray();
	DatagramPacket dp=new DatagramPacket(bytes,bytes.length,
					     ((InetWithPort)e.dest).host,
					     ((InetWithPort)e.dest).port);
	try {
	    mySocket.send(dp);
	    if(debugOn && debugOutput!=null)
		debugOutput.println("UDP: "+dp.getLength()+" bytes datagram sent to " +
				    dp.getAddress().getHostAddress() +
				    " (port "+dp.getPort()+")");
	}
	catch(java.io.IOException ex) {
	    System.err.println("Couldn't send message to socket. I should "+
			       "do something smart in here");
	}
    }

    private void insertOnChannel(SendableEvent e) {
	/* Init the event and send it up */
	try {
	    e.init();
	    e.go();
	    if(debugOn && debugOutput!=null)
		debugOutput.println("UDP: New event has been sent "+
				    "to channel");
	}
	catch(AppiaEventException ex) {
	    switch(ex.type) {
	    case ex.UNKNOWNSESSION :
		System.err.println("An exception stating that this "+
				   "class does not belong to the "+
				   "channel was raised in "+
				   "UdpSimpleSession");
		break;
	    case ex.ATTRIBUTEMISSING :
		System.err.println("Missing attribute exception in "+
				   "UdpSimpleSession");
		break;
	    case ex.UNWANTEDEVENT :
		System.err.println("The peer send an Unwanted event in "+
				   "UdpSimpleSession");
		break;
	    case ex.NOTINITIALIZED :
		System.err.println("Event not initialized exception "+
				   "ocurred after initialization in "+
				   "UdpSimpleSession");
		break;
	    }
	}
    }
}
